/*
 * poky-draw.c
 * This file is part of poky-engine
 *
 * Copyright (C) 2006 - OpenedHand Ltd
 *
 * poky-engine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * poky-engine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with poky-engine; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */


#include "poky-draw.h"
#include "poky-style.h"

#include <stdio.h>
#include <math.h>
#include <string.h>

#define DEBUG(func) //g_printf ("%s: detail = '%s'; state = %d; x:%d; y:%d; w:%d; h:%d;\n", func, detail, state_type, x, y, width, height);
#define DETAIL(foo) (strcmp (foo, detail) == 0)

GtkStyleClass *parent_style_class;

static void
poky_rounded_rectangle (GdkWindow * window, GdkGC * gc, gint x, gint y,
			gint width, gint height, GtkPositionType gap)
{
  GdkColor aa;
  GdkGCValues gc_values;

  gdk_gc_get_values (gc, &gc_values);

  gdk_colormap_query_color (gdk_drawable_get_colormap (window),
			    gc_values.foreground.pixel,
			    &gc_values.foreground.pixel);

  /* Brighten the colour for fake AA. Need a better way to do this */
  aa.red = MIN (65535, gc_values.foreground.red * 1.8);
  aa.blue = MIN (65535, gc_values.foreground.blue * 1.8);
  aa.green = MIN (65535, gc_values.foreground.green * 1.8);

  /** draw corners **/
  /* top left */
  if (gap != GTK_POS_TOP && gap != GTK_POS_LEFT)
  {
    gdk_draw_point (window, gc, x + 1, y + 1);
    gdk_draw_point (window, gc, x + 1, y + 2);
    gdk_draw_point (window, gc, x + 2, y + 1);

    gdk_gc_set_rgb_fg_color (gc, &aa);
    gdk_draw_point (window, gc, x + 2, y);
    gdk_draw_point (window, gc, x + 2, y + 2);
    gdk_draw_point (window, gc, x, y + 2);
  }
  if (gap != GTK_POS_BOTTOM && gap != GTK_POS_LEFT)
  {
    /* bottom left */
    gdk_gc_set_rgb_fg_color (gc, &gc_values.foreground);
    gdk_draw_point (window, gc, x + 1, y + height - 3);
    gdk_draw_point (window, gc, x + 1, y + height - 2);
    gdk_draw_point (window, gc, x + 2, y + height - 2);

    gdk_gc_set_rgb_fg_color (gc, &aa);
    gdk_draw_point (window, gc, x + 2, y + height - 1);
    gdk_draw_point (window, gc, x + 2, y + height - 3);
    gdk_draw_point (window, gc, x, y + height - 3);
  }
  if (gap != GTK_POS_TOP && gap != GTK_POS_RIGHT)
  {
    /* top right */
    gdk_gc_set_rgb_fg_color (gc, &gc_values.foreground);
    gdk_draw_point (window, gc, x + width - 3, y + 1);
    gdk_draw_point (window, gc, x + width - 2, y + 1);
    gdk_draw_point (window, gc, x + width - 2, y + 2);

    gdk_gc_set_rgb_fg_color (gc, &aa);
    gdk_draw_point (window, gc, x + width - 3, y);
    gdk_draw_point (window, gc, x + width - 3, y + 2);
    gdk_draw_point (window, gc, x + width - 1, y + 2);
  }
  if (gap != GTK_POS_BOTTOM && gap != GTK_POS_RIGHT)
  {
    /* bottom right */
    gdk_gc_set_rgb_fg_color (gc, &gc_values.foreground);
    gdk_draw_point (window, gc, x + width - 2, y + height - 3);
    gdk_draw_point (window, gc, x + width - 2, y + height - 2);
    gdk_draw_point (window, gc, x + width - 3, y + height - 2);

    gdk_gc_set_rgb_fg_color (gc, &aa);
    gdk_draw_point (window, gc, x + width - 3, y + height - 1);
    gdk_draw_point (window, gc, x + width - 3, y + height - 3);
    gdk_draw_point (window, gc, x + width - 1, y + height - 3);
  }

  /** draw lines **/
  gdk_gc_set_rgb_fg_color (gc, &gc_values.foreground);

  switch (gap)
  {
    case GTK_POS_TOP:
      y -= 3;
    case GTK_POS_BOTTOM:
      height += 3;
      break;

    case GTK_POS_LEFT:
      x -= 3;
    case GTK_POS_RIGHT:
      width += 3;
      break;
  }

  /* top and bottom */
  y++;
  height++;
  if (gap != GTK_POS_TOP)
    gdk_draw_line (window, gc, x + 3, y, x + width - 3, y);
  if (gap != GTK_POS_BOTTOM)
    gdk_draw_line (window, gc, x + 3, y + height - 3, x + width - 3,
		   y + height - 3);
  y--;
  height--;

  /* left and right */
  x++;
  width++;
  if (gap != GTK_POS_LEFT)
    gdk_draw_line (window, gc, x, y + 3, x, y + height - 3);
  if (gap != GTK_POS_RIGHT)
    gdk_draw_line (window, gc, x + width - 3, y + 3, x + width - 3,
		   y + height - 3);


}


static void
poky_gradient (GtkStyle * style, GdkWindow * window, GtkStateType state_type,
	       gint x, gint y, gint width, gint height)
{
  gint i, rd, gd, bd;		/* rd, gd, bd - change in r g and b for gradient */
  GdkColor start_color, end_color;
  GdkGC *gc;
  gc = gdk_gc_new (window);
  //gdk_gc_copy (gc, style->text_gc[state_type]);

  /* get the start and end colours */
  if (state_type == GTK_STATE_ACTIVE)
  {
    gdk_color_parse ("#73d216", &end_color);
    gdk_color_parse ("#4e9a06", &start_color);
  }
  else
  {
    gdk_color_parse ("#73d216", &start_color);
    gdk_color_parse ("#4e9a06", &end_color);
  }

  /* set line for 1px */
  gdk_gc_set_line_attributes (gc, 1, GDK_LINE_SOLID, GDK_CAP_BUTT,
			      GDK_JOIN_MITER);


  /* calculate the delta values */

  rd = (start_color.red - end_color.red) / MAX (height, 1);
  gd = (start_color.green - end_color.green) / MAX (height, 1);
  bd = (start_color.blue - end_color.blue) / MAX (height, 1);

  i = -1;
  while (++i < height)
  {
    gdk_gc_set_rgb_fg_color (gc, &start_color);
    gdk_draw_line (window, gc, x, y + i, x + width, y + i);
    start_color.red -= rd;
    start_color.blue -= bd;
    start_color.green -= gd;
  }

  /* round off the gradient corners */

  gdk_draw_point (window, style->bg_gc[state_type], x, y);
  gdk_draw_point (window, style->bg_gc[state_type], x, y + height - 1);
  gdk_draw_point (window, style->bg_gc[state_type], x + width - 1,
		  y + height - 1);
  gdk_draw_point (window, style->bg_gc[state_type], x + width - 1, y);

}

static void
poky_draw_box (DRAW_ARGS)
{

  DEBUG ("draw_box");

  /* initialise the background first */
  gdk_draw_rectangle (window, style->white_gc, TRUE, x, y, width, height);

  /*** draw the border ***/
  if (!DETAIL ("bar"))
    poky_rounded_rectangle (window, style->fg_gc[state_type], x, y, width,
			    height, -1);


  /*** draw the gradient ***/

  if (DETAIL ("trough") || DETAIL ("hscrollbar") || DETAIL ("vscrollbar"))
    return;

//  if (width < 20 || height < 20)
  //  return; /* don't bother with gradient on really small things */

  if (!DETAIL ("bar"))
  {
    x += 4;
    width -= 8;
    y += 4;
    height -= 8;
  }

  poky_gradient (style, window, state_type, x, y, width, height);

}

static void
poky_draw_shadow (DRAW_ARGS)
{

  DEBUG ("draw_shadow");


  if (widget && GTK_WIDGET_HAS_FOCUS (widget))
  {
    poky_rounded_rectangle (window, style->base_gc[GTK_STATE_SELECTED], x,
			    y, width, height, -1);
  }
  else
    poky_rounded_rectangle (window, style->fg_gc[state_type], x, y, width,
			    height, -1);

}

static void
poky_draw_check (GtkStyle * style, GdkWindow * window,
		 GtkStateType state_type, GtkShadowType shadow_type,
		 GdkRectangle * area, GtkWidget * widget,
		 const gchar * detail, gint x, gint y, gint width,
		 gint height)
{
  GdkGC *temp_gc = NULL;
  GdkColor color;

  DEBUG ("draw_shadow");

  temp_gc = gdk_gc_new (window);
  gdk_gc_copy (temp_gc, style->fg_gc[state_type]);
  gdk_gc_set_line_attributes (temp_gc, 2, GDK_LINE_SOLID, GDK_CAP_BUTT,
			      GDK_JOIN_MITER);

  if (shadow_type == GTK_SHADOW_IN)
    gdk_draw_rectangle (window, style->base_gc[GTK_STATE_SELECTED], TRUE,
			x + 2, y + 2, width - 4, height - 4);

  poky_rounded_rectangle (window, temp_gc, x, y, width, height, -1);

  g_object_unref (temp_gc);
}

static void
poky_draw_option (GtkStyle * style, GdkWindow * window,
		  GtkStateType state_type, GtkShadowType shadow_type,
		  GdkRectangle * area, GtkWidget * widget,
		  const gchar * detail, gint x, gint y, gint width,
		  gint height)
{
  GdkGC *temp_gc = NULL;

  temp_gc = gdk_gc_new (window);
  gdk_gc_copy (temp_gc, style->fg_gc[state_type]);
  gdk_gc_set_line_attributes (temp_gc, 2, GDK_LINE_SOLID, GDK_CAP_BUTT,
			      GDK_JOIN_MITER);

  if (shadow_type == GTK_SHADOW_IN)
    gdk_draw_arc (window, style->base_gc[GTK_STATE_SELECTED], TRUE, x, y,
		  width, height, 0, 360 * 64);
  gdk_draw_arc (window, temp_gc, FALSE, x, y, width, height, 0, 360 * 64);

  g_object_unref (temp_gc);

}

static void
poky_draw_box_gap (GtkStyle * style, GdkWindow * window,
		   GtkStateType state_type, GtkShadowType shadow_type,
		   GdkRectangle * area, GtkWidget * widget, gchar * detail,
		   gint x, gint y, gint width, gint height,
		   GtkPositionType gap_side, gint gap_x, gint gap_width)
{
  GdkRectangle rect;

  poky_rounded_rectangle (window, style->fg_gc[state_type], x, y, width, height, -1);


  switch (gap_side)
  {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
  }

  gtk_style_apply_default_background (style, window, TRUE, state_type, area,
				      rect.x, rect.y, rect.width,
				      rect.height);
}


static void
poky_draw_extension (GtkStyle * style, GdkWindow * window,
		     GtkStateType state_type, GtkShadowType shadow_type,
		     GdkRectangle * area, GtkWidget * widget, gchar * detail,
		     gint x, gint y, gint width, gint height,
		     GtkPositionType gap_side)
{

  switch (gap_side)
  {
    case GTK_POS_TOP:
      y -= 2;			/* bottom tab */
    case GTK_POS_BOTTOM:
      height += 2;
      break;			/* top tab */
    case GTK_POS_LEFT:
      x -= 2;			/* right tab */
    case GTK_POS_RIGHT:
      width += 2;
      break;			/* left tab */
  }

  poky_rounded_rectangle (window, style->fg_gc[state_type], x, y, width, height, gap_side);
  if (state_type == GTK_STATE_NORMAL)
    poky_gradient (style, window, state_type, x + 4, y + 4, width - 8,
		   height - 8);
}

void
poky_realise (GtkStyle * style)
{
  gint i;
  parent_style_class->realize (style);

  for (i = 0; i < 5; i += 1)
  {
    gdk_gc_set_line_attributes (style->fg_gc[i], 2, GDK_LINE_SOLID,
				GDK_CAP_BUTT, GDK_JOIN_MITER);
    gdk_gc_set_line_attributes (style->bg_gc[i], 2, GDK_LINE_SOLID,
				GDK_CAP_BUTT, GDK_JOIN_MITER);
    gdk_gc_set_line_attributes (style->text_gc[i], 2, GDK_LINE_SOLID,
				GDK_CAP_BUTT, GDK_JOIN_MITER);
    gdk_gc_set_line_attributes (style->base_gc[i], 2, GDK_LINE_SOLID,
				GDK_CAP_BUTT, GDK_JOIN_MITER);
  }
}


void
poky_draw_style_class_init (GtkStyleClass * style_class)
{

  parent_style_class = g_type_class_peek_parent (style_class);

  style_class->draw_shadow = poky_draw_shadow;
  style_class->draw_box = poky_draw_box;
  style_class->draw_check = poky_draw_check;
  style_class->draw_option = poky_draw_option;
  style_class->draw_box_gap = poky_draw_box_gap;
  style_class->draw_shadow_gap = poky_draw_box_gap;
  style_class->draw_extension = poky_draw_extension;
  style_class->realize = poky_realise;
/*
  style_class->draw_hline = draw_hline;
  style_class->draw_vline = draw_vline;
  style_class->draw_arrow = draw_arrow;
  style_class->draw_check = draw_check;
  style_class->draw_option = draw_option;
  style_class->draw_tab = draw_tab;
  style_class->draw_shadow_gap = draw_shadow_gap;
  style_class->draw_box_gap = draw_box_gap;
  style_class->draw_extension = draw_extension;
  style_class->draw_focus = draw_focus;
  style_class->draw_slider = draw_slider;
  style_class->draw_handle = draw_handle;
  style_class->draw_layout = draw_layout;
*/
}
